import re
from django.db import models
from pyvalem.formula import Formula
from pyvalem.stateful_species import StatefulSpecies

from _utils.models import QualifiedIDMixin


class Species(QualifiedIDMixin, models.Model):
    qid_prefix = "F"

    id = models.AutoField(primary_key=True)

    text = models.CharField(max_length=80)
    html = models.CharField(max_length=200)
    charge = models.SmallIntegerField(default=0, null=True)

    class Meta:
        verbose_name_plural = "Species"

    def __str__(self):
        return self.text

    @classmethod
    def get_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text. Uses pyvalem Formula.__repr__ for the canonicalisation.
        If not present, Species.DoesNotExist is raised.

        Parameters
        ----------
        text : str

        Returns
        -------
        Species
        """
        text_can = repr(Formula(text))
        return cls.objects.get(text=text_can)

    @classmethod
    def get_or_create_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text.

        Parameters
        ----------
        text : str

        Returns
        -------
        (Species, bool)
        """
        pyvalem_formula = Formula(text)
        text_can = repr(pyvalem_formula)
        try:
            return cls.objects.get(text=text_can), False
        except cls.DoesNotExist:
            # re-instantiate the pyvalem_formula with canonicalised
            # text to canonicalise html also:
            pyvalem_formula = Formula(text_can)
            species = cls.objects.create(
                text=text_can, charge=pyvalem_formula.charge, html=pyvalem_formula.html
            )
            return species, True


class SpeciesAlias(models.Model):
    text = models.CharField(max_length=80, unique=True)
    species = models.ForeignKey(Species, on_delete=models.CASCADE)

    def __str__(self):
        return f"{self.text} -> {self.species.text}"


class RP(QualifiedIDMixin, models.Model):
    qid_prefix = "RP"

    id = models.AutoField(primary_key=True)
    species = models.ForeignKey(Species, on_delete=models.CASCADE)

    text = models.CharField(max_length=200)
    html = models.CharField(max_length=600)

    def __str__(self):
        return self.text

    @classmethod
    def get_from_text(cls, text):
        """Looks for RP with equivalent canonicalised version of the
        text. Uses pyvalem StatefulSpecies.__repr__ for the canonicalisation.
        If not present, RP.DoesNotExist is raised.

        Parameters
        ----------
        text : str

        Returns
        -------
        RP
        """
        text_can = repr(StatefulSpecies(text))
        # TODO Look up Species formula in SpeciesAlias table.
        return cls.objects.get(text=text_can)

    @classmethod
    def filter_from_text(cls, text, inchi_lookup=False):
        """Filters for RP using canonicalised version of the StatefulSpecies
        represented by text, having first resolved the Species formula into
        its canonical form for this database by looking it up in the
        SpeciesAlias table.

        Parameters
        ----------
        text : str

        Returns
        -------
        django.db.models.query.QuerySet
        """

        patt = "[A-Z]{14}-[A-Z]{10}-N"
        if text.startswith("InChI=") or text.startswith("1S") or re.match(patt, text):
            try:
                species = SpeciesAlias.objects.get(text=text).species
                return cls.objects.filter(species__text=species.text)
            except SpeciesAlias.DoesNotExist:
                return []

        ss = StatefulSpecies(text)

        try:
            species = SpeciesAlias.objects.get(text=ss.formula).species
            rps = cls.objects.filter(species__text=species.text)
        except SpeciesAlias.DoesNotExist:
            rps = cls.objects.filter(species__text=ss.formula)

        for state in ss.states:
            rps = rps.filter(state__text=state)

        return rps

    @classmethod
    def get_or_create_from_text(cls, text):
        """Looks for a Species with equivalent canonicalised version of the
        text.

        Parameters
        ----------
        text : str

        Returns
        -------
        (RP, bool)
        """
        pyvalem_stateful_species = StatefulSpecies(text)
        text_can = repr(pyvalem_stateful_species)
        try:
            return cls.objects.get(text=text_can), False
        except cls.DoesNotExist:
            pyvalem_formula = pyvalem_stateful_species.formula
            species, _ = Species.get_or_create_from_text(repr(pyvalem_formula))
            # re-instantiate the pyvalem_stateful_species with canonicalised
            # text to canonicalise html also and sort the states consistently
            # with the text and html:
            pyvalem_stateful_species = StatefulSpecies(text_can)
            # build the RP instance:
            rp = cls.objects.create(
                species=species, text=text_can, html=pyvalem_stateful_species.html
            )
            # attach the states:
            for pyvalem_state in pyvalem_stateful_species.states:
                State.objects.create(
                    rp=rp,
                    text=repr(pyvalem_state),
                    html=pyvalem_state.html,
                    state_type=State.STATE_TYPE_MAP[pyvalem_state.__class__.__name__],
                )
            return rp, True

    @property
    def charge(self):
        """Returns the charge of RPs Species."""
        return self.species.charge


class State(QualifiedIDMixin, models.Model):
    qid_prefix = "S"

    KEY_VALUE_PAIR = 0
    GENERIC_EXCITED_STATE = 1
    ATOMIC_CONFIGURATION = 2
    ATOMIC_TERM_SYMBOL = 3
    DIATOMIC_MOLECULAR_CONFIGURATION = 4
    MOLECULAR_TERM_SYMBOL = 5
    VIBRATIONAL_STATE = 6
    ROTATIONAL_STATE = 10
    RACAH_SYMBOL = 11

    STATE_TYPE_CHOICES = (
        (KEY_VALUE_PAIR, "KeyValuePair"),
        (GENERIC_EXCITED_STATE, "GenericExcitedState"),
        (ATOMIC_CONFIGURATION, "AtomicConfiguration"),
        (ATOMIC_TERM_SYMBOL, "AtomicTermSymbol"),
        (DIATOMIC_MOLECULAR_CONFIGURATION, "DiatomicMolecularConfiguration"),
        (MOLECULAR_TERM_SYMBOL, "MolecularTermSymbol"),
        (VIBRATIONAL_STATE, "VibrationalState"),
        (ROTATIONAL_STATE, "RotationalState"),
        (RACAH_SYMBOL, "RacahSymbol"),
    )

    STATE_TYPE_MAP = {v: k for k, v in STATE_TYPE_CHOICES}

    id = models.AutoField(primary_key=True)
    rp = models.ForeignKey(RP, on_delete=models.CASCADE)

    state_type = models.SmallIntegerField(choices=STATE_TYPE_CHOICES)
    text = models.CharField(max_length=64)
    html = models.CharField(max_length=100)

    def __str__(self):
        return self.text



from django.test import TestCase

from rxn.models import ProcessType


class TestProcessType(TestCase):
    def test_str(self):
        process_type = ProcessType.objects.create(
            abbreviation="HDS",
            description="Dissociation",
            example_html="A + BC → A + B + C",
        )
        self.assertEqual(str(process_type), "HDS")

    def test_repr(self):
        process_type = ProcessType.objects.create(
            id=42,
            abbreviation="HDS",
            description="Dissociation",
            example_html="A + BC → A + B + C",
        )
        self.assertEqual(repr(process_type), "<P42: HDS>")



import warnings

from django.test import TestCase
from pyvalem.reaction import Reaction as PVReaction

from rp.models import RP, Species, State
from rxn.models import Reaction, ProcessType, ReactantList, ProductList
from pyvalem.reaction import ReactionParseError


class TestReaction(TestCase):
    def setUp(self):
        self.pt_hds = ProcessType.objects.create(
            abbreviation="HDS",
            description="Dissociation",
            example_html="A + BC → A + B + C",
        )
        self.pt_oth = ProcessType.objects.create(
            abbreviation="___", description="", example_html=""
        )
        ProcessType.objects.create(
            abbreviation="ENI",
            description="Negative Ion Formation",
            example_html="e- + A → A-",
        )
        ProcessType.objects.create(
            abbreviation="EDS",
            description="Electron-Impact Dissociation",
            example_html="e- + AB → A + B + e-",
        )

    def test_all_from_text_single(self):
        r_str = "H + H + He -> H + H **;n=2 + He * n=0"
        self.assertEqual(len(Reaction.all_from_text(r_str)), 0)
        self.assertEqual(len(RP.objects.all()), 0)

        pyvalem_reaction = PVReaction(r_str)

        r = Reaction.objects.create(
            text=repr(pyvalem_reaction), html=pyvalem_reaction.html
        )
        for _, pyvalem_stateful_species in pyvalem_reaction.reactants:
            ReactantList.objects.create(
                reaction=r,
                rp=RP.get_or_create_from_text(repr(pyvalem_stateful_species))[0],
            )
        for _, pyvalem_stateful_species in pyvalem_reaction.products:
            ProductList.objects.create(
                reaction=r,
                rp=RP.get_or_create_from_text(repr(pyvalem_stateful_species))[0],
            )
        r.process_types.add(self.pt_oth)

        for r_text in [
            "2H + He -> H + H **;n=2 + He *;n=0",
            "H + H + He -> H + H **;n=2 + He *;n=0",
            "2H + He -> H + H n=2, ** + He *;n=0",
            "2H + He -> H + H n=2;** + He n=0;*",
            "H + H + He -> H + H n=2;** + He n=0;*",
        ]:
            with self.subTest(r_text=r_text):
                self.assertEqual(len(Reaction.all_from_text(r_text)), 1)

        self.assertEqual(len(RP.objects.all()), 4)
        self.assertEqual(len(Species.objects.all()), 2)

    def test_all_from_text_multiple(self):
        Reaction.get_or_create_from_text("He + H + H -> 2H + He *;n=2", "c1")
        Reaction.get_or_create_from_text("He + 2H -> H + H + He *;n=2", "c2")
        Reaction.get_or_create_from_text(
            "He + 2H -> 2H + He *;n=2", "c2", process_type_abbreviations=("EDS",)
        )
        Reaction.get_or_create_from_text(
            "He + 2H -> 2H + He n=2;*", process_type_abbreviations=("EDS",)
        )
        self.assertEqual(len(Reaction.all_from_text("He + 2H -> 2H + He n=2;*")), 4)
        self.assertEqual(len(Reaction.objects.all()), 4)

    def test_all_from_text_non_existent(self):
        self.assertEqual(len(Reaction.all_from_text("H + H -> H + H")), 0)

    def test_does_not_create_duplicates(self):
        Reaction.get_or_create_from_text("H + H -> H + H", comment="foo")
        _, created = Reaction.get_or_create_from_text("2H -> 2H", comment="foo")
        self.assertFalse(created)
        self.assertEqual(len(RP.objects.all()), 1)

    def test_create_non_existing_rps(self):
        RP.get_or_create_from_text("He *;n=2")
        self.assertEqual(len(RP.objects.all()), 1)
        self.assertEqual(len(State.objects.all()), 2)
        self.assertEqual(len(Species.objects.all()), 1)
        self.assertEqual(len(Reaction.objects.all()), 0)

        Reaction.get_or_create_from_text("He n=2;* + 2H -> H + H + He n=2;*")
        self.assertEqual(len(RP.objects.all()), 2)
        self.assertEqual(len(State.objects.all()), 2)
        self.assertEqual(len(Species.objects.all()), 2)
        self.assertEqual(len(Reaction.objects.all()), 1)

    def test_get_or_create_from_text(self):
        self.assertEqual(Reaction.objects.count(), 0)
        self.assertEqual(Species.objects.count(), 0)
        self.assertEqual(RP.objects.count(), 0)
        self.assertEqual(State.objects.count(), 0)

        r1, created = Reaction.get_or_create_from_text(
            "H2 + e- -> H + H-", comment="the first"
        )
        self.assertEqual(str(r1), "e- + H2 → H + H-")
        self.assertTrue(created)
        self.assertEqual(Reaction.objects.count(), 1)  # created
        r2, created = Reaction.get_or_create_from_text(
            "e- + H2 -> H + H-", comment="the first"
        )
        self.assertEqual(r1, r2)
        self.assertFalse(created)
        self.assertEqual(Reaction.objects.count(), 1)  # not created
        r3, created = Reaction.get_or_create_from_text(
            "H2 + e- -> H + H-", comment="the second"
        )
        self.assertNotEqual(r1, r3)
        self.assertTrue(created)
        self.assertEqual(Reaction.objects.count(), 2)  # created

        r1, created = Reaction.get_or_create_from_text(
            "H2 + e- -> H + H-", process_type_abbreviations=("EDS",)
        )
        self.assertEqual(str(r1), "e- + H2 → H + H-")
        self.assertTrue(created)
        self.assertEqual(Reaction.objects.count(), 3)  # created
        r2, created = Reaction.get_or_create_from_text(
            "H2 + e- -> H + H-", process_type_abbreviations=("EDS",)
        )
        self.assertEqual(str(r2), "e- + H2 → H + H-")
        self.assertEqual(r1, r2)
        self.assertFalse(created)
        self.assertEqual(Reaction.objects.count(), 3)  # not created
        r3, created = Reaction.get_or_create_from_text(
            "e- + H2 -> H + H-", process_type_abbreviations=("EDS", "ENI")
        )
        self.assertEqual(str(r3), "e- + H2 → H + H-")
        self.assertNotEqual(r1, r3)
        self.assertTrue(created)
        self.assertEqual(Reaction.objects.count(), 4)  # created

        # the whole thing should have created 4 RPs, 4 Species, 0 States
        self.assertEqual(Species.objects.count(), 4)
        self.assertEqual(RP.objects.count(), 4)
        self.assertEqual(State.objects.count(), 0)

    def test_get_or_create_from_text_ordered(self):
        self.assertEqual(Reaction.objects.count(), 0)
        self.assertEqual(Species.objects.count(), 0)
        self.assertEqual(RP.objects.count(), 0)
        self.assertEqual(State.objects.count(), 0)

        r1, created = Reaction.get_or_create_from_text(
            "H 1s + H+ -> H+ + H+ + e-", comment="atom first"
        )
        r2, created = Reaction.get_or_create_from_text(
            "H+ + H 1s -> H+ + H+ + e-", comment="ion first"
        )
        self.assertEqual(r1.ordered_text, r2.ordered_text)

        r3, created = Reaction.get_or_create_from_text(
            "H2 + e- + H2 + e- <=> e- + H4+2 + 2e- + e-", comment="test 1"
        )
        r4, created = Reaction.get_or_create_from_text(
            "e- + 2H2 + e- <=> H4+2 + 4e-", comment="test 1"
        )
        self.assertEqual(r3.ordered_text, r4.ordered_text)
        self.assertEqual(r3.ordered_text, "2e- + H2 + H2 ⇌ 4e- + H4+2")

        r5, created = Reaction.get_or_create_from_text(
            "H+ + Li 2s -> Li+", strict=False
        )
        r6, created = Reaction.get_or_create_from_text(
            "Li 2s + H+ -> Li+", strict=False
        )
        self.assertEqual(r5.ordered_text, r6.ordered_text)
        self.assertEqual(r5.ordered_text, "H+ + Li 2s → Li+")

    def test_nonstrict_reactions(self):
        s_r = "Li + e- -> Li+"

        self.assertRaises(ReactionParseError, Reaction.get_or_create_from_text, s_r)

        r, _ = Reaction.get_or_create_from_text(s_r, strict=False)
        self.assertEqual(repr(r), f"<R{r.id}: e- + Li → Li+>")

    def test_molecularity(self):
        reaction, _ = Reaction.get_or_create_from_text("5H + 5e- -> H- + H- + 3H-")
        self.assertEqual(reaction.molecularity, 10)
        self.assertEqual(len(RP.objects.all()), 3)
        self.assertEqual(len(State.objects.all()), 0)
        self.assertEqual(len(Species.objects.all()), 3)
        self.assertEqual(len(Reaction.objects.all()), 1)

    def test_process_types(self):
        r, _ = Reaction.get_or_create_from_text(
            "H + H -> H + H", process_type_abbreviations=("___",)
        )
        self.assertEqual(list(r.process_types.all()), [self.pt_oth])
        self.assertEqual(list(self.pt_oth.reaction_set.all()), [r])

    def test_comment(self):
        r, _ = Reaction.get_or_create_from_text("H + H -> H + H", comment="foo")
        self.assertEqual(r.comment, "foo")

    def test_str(self):
        r, _ = Reaction.get_or_create_from_text("2H -> 2H", comment="foo")
        self.assertEqual(str(r), "H + H → H + H")

    def test_repr(self):
        r, _ = Reaction.get_or_create_from_text("2H -> H + H", comment="foo")
        self.assertEqual(repr(r), f"<R{r.id}: H + H → H + H>")

    def test_reset_html(self):
        r, _ = Reaction.get_or_create_from_text("Be+4 + H 2p1 m=1 → Be+4 + H *")
        self.assertEqual(repr(r), f"<R{r.id}: Be+4 + H 2p;m=1 → Be+4 + H *>")
        self.assertEqual(
            r.html, "Be<sup>4+</sup> + H 2p m=1 → Be<sup>4+</sup> + H <sup>*</sup>"
        )

        r.text = "Be+4 + H 2p1 |m|=1 → Be+4 + H *"
        r._reset_html()
        self.assertEqual(
            r.html, "Be<sup>4+</sup> + H 2p |m|=1 → Be<sup>4+</sup> + H <sup>*</sup>"
        )

    def test_latex(self):
        r, _ = Reaction.get_or_create_from_text(
            "e- + BeH+ X(1Σ+) v=0 -> BeH+ X(1Σ+) v=3 + e-"
        )
        self.assertEqual(
            r.latex,
            r"\mathrm{e}^- + \mathrm{Be}\mathrm{H}^{+} \; X{}^{1}\Sigma^+ \; v=0 \rightarrow \mathrm{Be}\mathrm{H}^{+} \; X{}^{1}\Sigma^+ \; v=3 + \mathrm{e}^-",
        )



from django.test import TestCase
from pyvalem.formula import FormulaParseError
from pyvalem.states import StateParseError
from pyvalem.stateful_species import StatefulSpeciesError

from rp.models import Species, RP, State


# noinspection PyTypeChecker
class TestRP(TestCase):
    valid_states = {
        "n=1": 0,
        "**": 1,
        "1s2.2s1": 2,
        "2Po_1/2": 3,
        "1σu2": 4,
        "1Σ-": 5,
        "v=1": 6,
        "J=3/2": 10,
        # what about the Racah symbol?
    }
    valid_species = {
        "T",
        "H2",
        "D2-",
        "C3H8+2",
        "(235U)",
        "(1H)2(16O)",
        "ortho-C6H4(CH3)2",
    }
    invalid_species = ["foo"]
    invalid_states = ["foo"]

    def setUp(self):
        self.test_species = Species.objects.create(text="H", html="H")
        self.test_state1_kwargs = {"state_type": 0, "text": "n=1", "html": "n=1"}
        self.test_state2_kwargs = {"state_type": 1, "text": "**", "html": "**"}

    def test_raw_instantiation(self):
        rp = RP.objects.create(species=self.test_species)
        self.assertIs(rp.species, self.test_species)
        self.assertEqual(len(rp.state_set.all()), 0)

    def test_raw_states_assign(self):
        rp = RP.objects.create(species=self.test_species)
        State(rp=rp, **self.test_state1_kwargs).save()
        self.assertEqual(len(rp.state_set.all()), 1)
        # duplicate states are allowed by the raw django API
        State(rp=rp, **self.test_state1_kwargs).save()
        self.assertEqual(len(rp.state_set.all()), 2)

    def test_get_from_text_raises(self):
        # getting non-existing RP instances raises DoesNotExist errors
        for rp_text in [
            f"{species} {state}"
            for species in self.valid_species
            for state in self.valid_states
        ]:
            with self.subTest(text=rp_text):
                # noinspection PyTypeChecker
                with self.assertRaises(RP.DoesNotExist):
                    RP.get_from_text(rp_text)

    def test_get_from_text_non_existing(self):
        # getting RP instance from string with either non-existing species, or
        # without all states, raises DoesNotExist
        rp = RP.objects.create(species=self.test_species)
        state1 = State.objects.create(rp=rp, **self.test_state1_kwargs)
        state2 = State.objects.create(rp=rp, **self.test_state2_kwargs)
        non_existing_rp_texts = [
            f"{rp.species.text}",  # no states
            f"{rp.species.text} {state1.text}",  # only the first state
            f"{rp.species.text} {state2.text}",  # only the second state
            f"W+42 {state1.text};{state2.text}",  # corr. states, non-exist. sp
        ]  # those are all valid pyvalem-parsable RP text examples
        for rp_text in non_existing_rp_texts:
            with self.subTest(rp_text=rp_text):
                # noinspection PyTypeChecker
                with self.assertRaises(RP.DoesNotExist):
                    RP.get_from_text(rp_text)

    def test_get_from_text_existing(self):
        # get_from_text should not care about the order of the states!
        RP.get_or_create_from_text("H2+ v=2;3SIGMA+g")
        existing_rp_texts = ["H2+ v=2;3SIGMA+g", "H2+ 3Σ+g; v=2"]
        for rp_text in existing_rp_texts:
            with self.subTest(rp_text=rp_text):
                RP.get_from_text(rp_text)

    def test_get_or_create_from_text_one_state(self):
        # RP creating
        rp_text = "H2+ *"
        rp, created = RP.get_or_create_from_text(rp_text)
        self.assertTrue(created)
        self.assertEqual(len(RP.objects.all()), 1)
        # check the RP instance:
        self.assertEqual(rp.species.text, "H2+")
        self.assertEqual(rp.species.html, "H<sub>2</sub><sup>+</sup>")
        self.assertEqual(rp.species.charge, 1)
        self.assertEqual(rp.state_set.get().text, "*")
        self.assertEqual(rp.state_set.get().html, "<sup>*</sup>")
        self.assertEqual(rp.state_set.get().state_type, State.GENERIC_EXCITED_STATE)

    def test_get_or_create_from_text_swap_states(self):
        # RP get_or_creating multiple times with swapped states will not result
        # in multiple RP instances.
        rp_text = "H2 *;n=1;v=1"
        self.assertEqual(len(RP.objects.all()), 0)
        rp, created = RP.get_or_create_from_text(rp_text)
        self.assertTrue(created)
        num_rps = 1
        self.assertEqual(len(RP.objects.all()), num_rps)
        num_species = len(Species.objects.all())
        num_states = len(State.objects.all())
        # check the RP instance:
        for states_combo in [
            "*;n=1;v=1",
            "*;v=1;n=1",
            "n=1;v=1;*",
            "n=1;*;v=1",
            "v=1;*;n=1",
            "v=1;n=1;*",
        ]:
            rp_new, created = RP.get_or_create_from_text(f"H2 {states_combo}")
            self.assertEqual(rp_new, rp)
            self.assertFalse(created)
        # no new species, states or rp instances created:
        self.assertEqual(len(RP.objects.all()), num_rps)
        self.assertEqual(len(State.objects.all()), num_states)
        self.assertEqual(len(Species.objects.all()), num_species)

    def test_get_or_create_from_text_existing_species(self):
        # The existing Species instance is not re-created by
        # RP.get_or_create_from_text
        species_text = "W+42"
        Species.objects.create(text=species_text)
        num_species = len(Species.objects.all())
        num_states = len(State.objects.all())
        # Create RP instance with existing species
        _, created = RP.get_or_create_from_text(f"{species_text} key=value")
        self.assertTrue(created)
        # No new species were created
        self.assertEqual(len(Species.objects.all()), num_species)
        # New states were created
        self.assertEqual(len(State.objects.all()), num_states + 1)

    def test_get_or_create_from_invalid_text(self):
        # Testing that couple of invalid text cases raise the appropriate
        # pyvalem exceptions.
        for species_text in self.invalid_species:
            for state_text in self.valid_states:
                with self.assertRaises(FormulaParseError):
                    RP.get_or_create_from_text(f"{species_text} {state_text}")
        for species_text in self.valid_species:
            for state_text in self.invalid_states:
                with self.assertRaises(StateParseError):
                    RP.get_or_create_from_text(f"{species_text} {state_text}")
        with self.assertRaises(StatefulSpeciesError):
            RP.get_or_create_from_text("H2 *;***")
        with self.assertRaises(StatefulSpeciesError):
            RP.get_or_create_from_text("H 1s1;2p1")

    def test_text_attribute(self):
        rp_text = "(235U) l=0;n=1;***"
        rp, _ = RP.get_or_create_from_text(rp_text)
        text = rp.text
        rp.delete()
        with self.assertRaises(RP.DoesNotExist):
            RP.get_from_text(rp_text)

        equivalent_text = "(235U) l=0;***;n=1"
        rp, created = RP.get_or_create_from_text(equivalent_text)
        self.assertTrue(created)
        text_equivalent = rp.text

        self.assertEqual(text, text_equivalent)

    def test_html_attribute(self):
        rp_text = "(235U) n=2;3P"
        rp, _ = RP.get_or_create_from_text(rp_text)
        html = rp.text
        rp.delete()
        with self.assertRaises(RP.DoesNotExist):
            RP.get_from_text(rp_text)

        equivalent_text = "(235U) 3P;n=2"
        rp, _ = RP.get_or_create_from_text(equivalent_text)
        html_equivalent = rp.text

        self.assertEqual(html, html_equivalent)

    def test_str(self):
        rp, _ = RP.get_or_create_from_text("H *")
        self.assertEqual(str(rp), rp.text)

    def test_repr(self):
        rp, _ = RP.get_or_create_from_text("H *")
        self.assertEqual(repr(rp), f"<RP{rp.pk}: H *>")

    def test_cascade_delete(self):
        # deletion of a Species will delete its RP
        RP.objects.create(species=self.test_species)
        num_species = len(Species.objects.all())
        num_rps = len(RP.objects.all())
        self.test_species.delete()
        # Of course Species was deleted:
        self.assertEqual(len(Species.objects.all()), num_species - 1)
        # RP was also deleted by the deletion cascade:
        self.assertEqual(len(RP.objects.all()), num_rps - 1)

    def test_cascade_delete_reversed(self):
        # deletion of an RP will not delete its Species
        rp = RP.objects.create(species=self.test_species)
        num_species = len(Species.objects.all())
        num_rps = len(RP.objects.all())
        rp.delete()
        self.assertEqual(len(Species.objects.all()), num_species)
        self.assertEqual(len(RP.objects.all()), num_rps - 1)

    def test_canonical_state_representation(self):
        # Existing species must be found even if I pass text with e.g. SIGMA
        # instead of Σ.
        for equivalent_strings, nstates in [
            (["H2 1sigma", "H2 1σ", "H2 1sigma1", "H2 1σ1"], 1),
            (["H2 1SIGMA-", "H2 1Σ-"], 1),
            (["H2 v=0 J=2", "H2 v=0;J=2", "H2 v=0; J=2", "H2 v=0,J=2"], 2),
            (["H2 J=3/2", "H2 J=1.5"], 1),
        ]:
            num_rps = len(RP.objects.all())
            num_states = len(State.objects.all())
            base_text = equivalent_strings[0]
            base_rp, created = RP.get_or_create_from_text(base_text)
            self.assertTrue(created)
            self.assertEqual(len(RP.objects.all()), num_rps + 1)
            self.assertEqual(len(State.objects.all()), num_states + nstates)
            for equivalent_text in equivalent_strings:
                equivalent_rp, created = RP.get_or_create_from_text(equivalent_text)
                with self.subTest(base_text=base_text, equivalent_text=equivalent_text):
                    # assert that the pulled RPs are the same
                    self.assertFalse(created)
                    self.assertEqual(equivalent_rp, base_rp)
            # assert that no new RPs and States were created
            self.assertEqual(len(RP.objects.all()), num_rps + 1)
            self.assertEqual(len(State.objects.all()), num_states + nstates)

    def test_charge(self):
        self.assertEqual(RP.get_or_create_from_text("H")[0].charge, 0)
        self.assertEqual(RP.get_or_create_from_text("H+")[0].charge, 1)
        self.assertEqual(RP.get_or_create_from_text("He+2")[0].charge, 2)
        self.assertEqual(RP.get_or_create_from_text("He-2")[0].charge, -2)



from django.test import TestCase
from refs.models import Ref

from rxn.models import Reaction, ProcessType
from .models import MyReactionDataSet


class TestReactionDataSet(TestCase):
    def setUp(self):
        process_types_raw = [
            ["EEX", "Excitation", "e<sup>-</sup> + A → e<sup>-</sup> + A<sup>*</sup>"],
            [
                "EXV",
                "Vibrational Excitation",
                "e<sup>-</sup> + AB → e<sup>-</sup> + AB (<i>v</i>=*)",
            ],
        ]
        self.all_process_types = {}
        for abbreviation, description, example_html in process_types_raw:
            pt = ProcessType.objects.create(
                abbreviation=abbreviation,
                description=description,
                example_html=example_html,
            )
            self.all_process_types[abbreviation] = pt

        self.test_reaction, _ = Reaction.get_or_create_from_text(
            "BeH+ v=0 + e- -> BeH+ v=10 + e-", process_type_abbreviations=("EEX", "EXV")
        )

        self.doi = "10.1016/j.adt.2016.09.002"
        Ref.objects.create(
            authors="S. Niyonzima et.al.",
            title="Low-energy collisions between electrons and BeH+",
            journal="Atomic Data and Nuclear Data Tables",
            volume="115-116",
            page_start="287",
            page_end="308",
            year=2017,
            doi=self.doi,
        )

    def test_add_dataset(self):
        json_data = '{"T_max": {"value": 5000, "units": "K"}}'
        json_comment = '{"comment": "comment"}'
        dataset = MyReactionDataSet.objects.create(
            reaction=self.test_reaction, json_comment=json_comment, json_data=json_data
        )
        ref = Ref.objects.get(doi=self.doi)
        dataset.refs.add(ref)

    def test_str(self):
        ds = MyReactionDataSet.objects.create(reaction=self.test_reaction)
        self.assertEqual(str(ds), str(self.test_reaction))

    def test_repr(self):
        ds = MyReactionDataSet.objects.create(id=42, reaction=self.test_reaction)
        self.assertEqual(repr(ds), f"<D42: {str(self.test_reaction)}>")



from django.core.exceptions import ValidationError
from django.test import TestCase

from rp.models import State, Species, RP


class TestState(TestCase):
    def setUp(self):
        self.test_species = Species.objects.create(text="H", html="H")
        self.test_rp = RP.objects.create(species=self.test_species)
        self.test_state_kwargs = {
            "rp": self.test_rp,
            "state_type": 0,
            "text": "n=1",
            "html": "n=1",
        }

    def test_instantiation(self):
        State(**self.test_state_kwargs)
        State()

    def test_instance_creation_succeeds(self):
        self.assertEqual(len(State.objects.all()), 0)
        state = State.objects.create(**self.test_state_kwargs)
        self.assertEqual(len(State.objects.all()), 1)
        for kwarg, val in self.test_state_kwargs.items():
            self.assertEqual(getattr(State.objects.get(pk=state.pk), kwarg), val)
        self.assertIs(self.test_rp, state.rp)

    def test_unsupported_state_type(self):
        unsupported_state_types = [-1, 15]
        supported_state_types = [0, 1]
        for state_type in unsupported_state_types:
            kwargs = dict(self.test_state_kwargs, state_type=state_type)
            state = State(**kwargs)
            with self.subTest(state_type=state_type):
                with self.assertRaises(ValidationError):
                    state.clean_fields()
        for state_type in supported_state_types:
            kwargs = dict(self.test_state_kwargs, state_type=state_type)
            state = State(**kwargs)
            state.clean_fields()
            state.save()
            self.assertEqual(State.objects.get(pk=state.pk).state_type, state_type)
        self.assertEqual(len(State.objects.all()), len(supported_state_types))

    def test_str(self):
        self.assertEqual(
            str(State.objects.create(**self.test_state_kwargs)),
            self.test_state_kwargs["text"],
        )

    def test_repr(self):
        st = State.objects.create(**self.test_state_kwargs)
        self.assertEqual(repr(st), f"<S{st.pk}: n=1>")

    def test_cascade_delete(self):
        # deletion of an RP will delete all its States
        states_created = 3
        for _ in range(states_created):
            State.objects.create(**self.test_state_kwargs)
        num_rps = len(RP.objects.all())
        num_states = len(State.objects.all())
        self.test_rp.delete()
        self.assertEqual(len(RP.objects.all()), num_rps - 1)
        self.assertEqual(len(State.objects.all()), num_states - states_created)

    def test_cascade_delete_reversed(self):
        # deletion of a State will NOT delete it's RP
        state = State.objects.create(**self.test_state_kwargs)
        num_rps = len(RP.objects.all())
        num_states = len(State.objects.all())
        state.delete()
        self.assertEqual(len(State.objects.all()), num_states - 1)
        self.assertEqual(len(RP.objects.all()), num_rps)



from django.test import TestCase
from django.db.utils import IntegrityError
from rp.models import Species, SpeciesAlias, RP


# noinspection PyTypeChecker
class TestSpeciesAlias(TestCase):
    def setUp(self):
        self.HD = Species.objects.create(text="HD")
        self.HD_aliases = ["DH", "H(2H)", "(2H)H", "(1H)(2H)", "(2H)(1H)"]
        for HD_alias in self.HD_aliases:
            SpeciesAlias.objects.create(species=self.HD, text=HD_alias)
        self.Xep34 = Species.objects.create(text="Xe+34")
        self.Xep34_inchi = SpeciesAlias.objects.create(
            species=self.Xep34, text="InChI=1S/Xe/q+34"
        )
        self.Xep34_inchi2 = SpeciesAlias.objects.create(
            species=self.Xep34, text="1S/Xe/q+34"
        )
        self.Xep34_inchikey = SpeciesAlias.objects.create(
            species=self.Xep34, text="CBYSBAHMMKPXKC-UHFFFAOYSA-N"
        )

    def test_HD_aliases(self):
        for HD_alias in self.HD_aliases:
            species = SpeciesAlias.objects.get(text=HD_alias).species
            self.assertEqual(species, self.HD)

            with self.assertRaises(SpeciesAlias.DoesNotExist):
                SpeciesAlias.objects.get(text="xx")

    def test_create_nonunique_alias(self):
        species = Species.objects.create(text="H(35Cl)")
        SpeciesAlias.objects.create(species=species, text="(1H)(35Cl)")
        with self.assertRaises(IntegrityError):
            SpeciesAlias.objects.create(species=species, text="(1H)(35Cl)")

    def test_filter_RP_from_text(self):
        rp1, _ = RP.get_or_create_from_text(text="HD v=0 X(1SIGMA+g)")
        rp2, _ = RP.get_or_create_from_text(text="HD v=1 X(1SIGMA+g)")

        for text in "HD", "HD X(1SIGMA+g)":
            rps = RP.filter_from_text(text)
            self.assertEqual(rps.count(), 2)
            self.assertEqual(rps[0], rp1)
            self.assertEqual(rps[1], rp2)

        rps = RP.filter_from_text("H(2H) v=0")
        self.assertEqual(rps.count(), 1)
        self.assertEqual(rps[0], rp1)

        rps = RP.filter_from_text("DH X(1SIGMA+g)")
        self.assertEqual(rps.count(), 2)
        self.assertEqual(rps[0], rp1)
        self.assertEqual(rps[1], rp2)

        rps = RP.filter_from_text("LiH")
        self.assertFalse(rps.exists())

    def test_inchi_aliases(self):
        rp1, _ = RP.get_or_create_from_text(text="Xe+34 1s2.2p3 3P_1")
        for inchi_alias in (
            self.Xep34,
            self.Xep34_inchi,
            self.Xep34_inchi2,
            self.Xep34_inchikey,
        ):
            rps = RP.filter_from_text(inchi_alias.text, inchi_lookup=True)
            self.assertEqual(rps.count(), 1)
            self.assertEqual(rps[0], rp1)



from django.db import transaction
from django.test import TestCase
from pyvalem.formula import Formula

from rp.models import Species


class TestSpecies(TestCase):
    default_kwargs = {"text": "H", "charge": 0, "html": "H"}
    formula_text_choices = "H2", "T3", "(CH2)C(CH2)+42"

    def test_instantiation_succeeds(self):
        Species(text="foo", charge=0, html="")
        Species(text="", html=r"<tag>foo</tag>")
        Species(text=42, charge="abc", html=(42, 42.0, "42"))
        # actually instantiation succeeds with any arguments,
        # even without arguments? Errors are raised during saving.
        Species()

    def test_instantiation_defaults(self):
        sp = Species(text="H", html="H")
        self.assertEqual(sp.charge, 0)

    def test_instantiation_does_not_create(self):
        Species(**self.default_kwargs)
        self.assertEqual(len(Species.objects.all()), 0)

    def test_django_save_duplicates(self):
        # simply creating the instances through Django API can make duplicates.
        Species(**self.default_kwargs).save()
        self.assertEqual(len(Species.objects.all()), 1)
        Species(**self.default_kwargs).save()
        self.assertEqual(len(Species.objects.all()), 2)
        Species.objects.create(**self.default_kwargs)
        self.assertEqual(len(Species.objects.all()), 3)

    def test_integrity_failures(self):
        for args in [
            (None, 0, "H"),
            ("H", 0, None),
            ("H", "foo", "H"),
            ("H", "42.0", "H"),
        ]:
            kwargs = dict(zip(("text", "charge", "html"), args))
            sp = Species(**kwargs)
            with self.subTest(**kwargs):
                with self.assertRaises(Exception):
                    # exception depends on the database backend!
                    with transaction.atomic():
                        sp.save()
        self.assertEqual(len(Species.objects.all()), 0)

    def test_integrity_typecasting(self):
        for arg, passed, saved in [
            ("text", 42, "42"),
            ("text", int, "<class 'int'>"),
            ("charge", "-42", -42),
            ("charge", " 42 ", 42),
        ]:
            sp = Species(**dict(self.default_kwargs, **{arg: passed}))
            sp.save()
            with self.subTest(arg=arg, passed=passed, saved=saved):
                # the types are not converted upon saving but they are,
                # when the instance is fetched from the database...
                self.assertEqual(getattr(sp, arg), passed)
                self.assertEqual(getattr(Species.objects.get(pk=sp.pk), arg), saved)

    def test_str(self):
        sp = Species(pk=1, **self.default_kwargs)
        self.assertEqual(str(sp), self.default_kwargs["text"])

    def test_repr(self):
        sp = Species(pk=1, **self.default_kwargs)
        self.assertEqual(repr(sp), f"<F{sp.pk}: {str(sp)}>")

    def test_get_from_text_not_found(self):
        for text in self.formula_text_choices:
            with self.subTest(formula_text=text):
                # noinspection PyTypeChecker
                with self.assertRaises(Species.DoesNotExist):
                    Species.get_from_text(text)

    def test_get_from_text_found(self):
        kwargs = {"text": "H2O", "charge": 42, "html": ""}
        Species(**kwargs).save()
        species = Species.get_from_text(kwargs["text"])
        for attr, val in kwargs.items():
            self.assertEqual(getattr(species, attr), val)

    def test_get_or_create_from_text(self):
        for text in self.formula_text_choices:
            initial_num_species = len(Species.objects.all())
            # noinspection PyTypeChecker
            with self.assertRaises(Species.DoesNotExist):
                # species does not exist yet
                Species.get_from_text(text)

            with self.subTest(formula_text=text):
                # create the species
                species, created = Species.get_or_create_from_text(text)
                self.assertTrue(created)
                formula = Formula(text)
                # how does the species look?
                self.assertEqual(species.text, text)
                self.assertEqual(species.charge, formula.charge)
                self.assertEqual(species.html, formula.html)
                # was the species created?
                self.assertEqual(len(Species.objects.all()), initial_num_species + 1)
                self.assertEqual(species, Species.get_from_text(text))
                # will the species not be created again?
                species1, created = Species.get_or_create_from_text(text)
                self.assertFalse(created)
                self.assertEqual(len(Species.objects.all()), initial_num_species + 1)
                self.assertEqual(species1, species)



# Generated by Django 3.0 on 2022-11-25 16:16

from django.db import migrations, models


class Migration(migrations.Migration):
    dependencies = [
        ("rxn", "0002_reaction_latex"),
    ]

    operations = [
        migrations.AddField(
            model_name="reaction",
            name="ordered_text",
            field=models.CharField(default="a", editable=False, max_length=256),
            preserve_default=False,
        ),
    ]
